__author__ = 'aymgal'

from copy import deepcopy
from typing import Tuple

from coolest.template.classes.likelihood import Likelihood, SUPPORTED_CHOICES
from coolest.template.classes.base import APIBaseObject

class DataLikelihoodList(list, APIBaseObject):

    """
    A list all likelihood terms used during lens model parameters optimization. 
    Only applicable in the 'MAP' mode.
    """

    def __init__(self, *likelihoods: Tuple[Likelihood]) -> None:
        # check that there is no duplicate likelihood type
        unique_choices = deepcopy(SUPPORTED_CHOICES)
        for ll in likelihoods:
            if ll.type in unique_choices:
                unique_choices.remove(ll.type)
            else:
                raise ValueError(f"There cannot be more than one likelihood type '{ll.type}'.")
        list.__init__(self, likelihoods)
        APIBaseObject.__init__(self)
    
    def index(self, likelihood_type):
        """Override the parent class to look for a likelihood from its string type.

        Parameters
        ----------
        likelihood_type : string
            Type of likelihood
        """
        found = False
        for i, ll in enumerate(self):
            if ll.type == likelihood_type:
                return i
        raise ValueError(f"Likelihood '{likelihood_type}' is not in the likelihood list.")

    def check_consistency_with_observation(self, observation):
        for ll in self:
            ll.check_consistency_with_observation(observation)
        


__author__ = 'aymgal'

import numpy as np

from coolest.template.classes.base import APIBaseObject
from coolest.template.classes.grid import PixelatedRegularGrid


__all__ = [
    'ImagingDataLikelihood',
]

SUPPORTED_CHOICES = [
    'ImagingDataLikelihood', 

    # TODO: implement the following classes:
    # 'TimeDelayLikelihood',
    # 'MagnificationRatioLikelihood',
    # 'ImagePlanePositionLikelihood',
    # 'SourcePlanePositionLikelihood',
]


class Likelihood(APIBaseObject):
    """Abstract likelihood object.

    Parameters
    ----------
    name : str
        Name of the entity (e.g., the name of the galaxy)
    redshift : float
        Redshift of the entity, if known / assumed
    mass_model : MassModel, optional
        Mass model (i.e., a set of mass profiles), by default None

    Raises
    ------
    ValueError
        If the provided redshift is negative.
    """

    def __init__(self, *args, **kwargs) -> None:
        self.type = self.__class__.__name__  # name of the children class
        super().__init__()


class ImagingDataLikelihood(Likelihood):
    """Imaging data likelihood, which typically holds the type of
    likelihood (e.g., Gaussian) and information about masked pixels.

    Parameters
    ----------
    mask : PixelatedRegularGrid, optional
        Binary mask with 0 for masked areas and 1 for areas included
        in the likelihood computation, by default None
    """

    def __init__(self, 
                 mask: PixelatedRegularGrid = None,
                ) -> None:
        self.mask = mask
        super().__init__()
    
    def get_mask_pixels(self, directory=None):
        pixels = self.mask.get_pixels(directory=directory)
        # checks that it contains only 0 and 1
        assert np.all(np.isin(pixels, [0, 1])), "Imaging likelihood is not binary."
        return pixels

    def check_consistency_with_observation(self, observation):
        """Checks that the data image is consistent with instrument properties"""
        width  = abs(self.mask.field_of_view_x[1] - self.mask.field_of_view_x[0])
        height = abs(self.mask.field_of_view_y[1] - self.mask.field_of_view_y[0])
        num_pix_ra = round(width / observation.pixels.pixel_size)
        error_message_ra = f"Number of pixels along x is inconsistent (data: {num_pix_ra}, likelihood mask: {self.mask.num_pix_x})."
        assert self.mask.num_pix_x  == num_pix_ra, error_message_ra
        num_pix_dec = round(height / observation.pixels.pixel_size)
        error_message_dec = f"Number of pixels along y is inconsistent (data: {num_pix_dec}, likelihood mask: {self.mask.num_pix_y})."
        assert self.mask.num_pix_y  == num_pix_dec, error_message_dec
        # TODO: check pixel size value?
